(** * Types and functins for matrices *)
Require Import List.
Require Import Setoid.
Require Import Bool.
Require Import Arith.
Require Import LIFO.Nat.BoundedNat.
Require Import StrongNth.
Require Import BoolFunctions.
Require Import Fold.
Require Import InSig.
Require Import Misc.
Require Import SeqNat.
Require Import Map.
Require Import Length.


Set Implicit Arguments.

(** * [isMatrix] a boolean function to caracterizz a list of lists as
   a matrix: either empty or all its elements have the same non zero
   length. *)

Definition isMatrix (A:Type) (m:list(list A)) :=
  match m with 
    | nil => true
    | h::t => 
      let n := List.length h in 
        negb (beq_nat n 0) && forallb (isLengthOf n) t
  end.

(** ** Facts on [isMatrix] *)

Lemma sameLengthsSuccIsMatrix:
  forall (A:Type) (m:list(list A))(n:nat),
  forallb (isLengthOf (S n)) m = true ->
  isMatrix m = true.
Proof.
  intros A m n H; destruct m. 
    auto. 
    simpl in * ; rewrite andb_true_iff in H; destruct H as [H1 H2];
      unfold isLengthOf in H1; apply beq_nat_true in H1. rewrite H1;
        assumption.
Qed.

Lemma sameLengthsNotZeroIsMatrix:
  forall (A:Type) (m:list(list A))(n:nat),
  negb(beq_nat n 0) && forallb (isLengthOf n) m = true ->
  isMatrix m = true.
Proof.
  intros A m n H; 
    rewrite andb_true_iff in H; destruct H as [H1 H2];
      case_eq(beq_nat n 0); intros H'; try(rewrite H' in H1 ; discriminate);
        destruct n ; try(discriminate); apply sameLengthsSuccIsMatrix with (n:=n);
          assumption.
Qed.

Lemma revOfMatrixIsMatrix:
  forall (A:Type) (m:list(list A)),
  isMatrix m = true -> isMatrix(rev m) = true.
Proof.
  intros A m H; induction m using rev_ind.
    auto.
    rewrite rev_unit ; destruct m.
      auto.
      simpl in *; rewrite forallbApp in H; rewrite andb_true_iff in H;
        destruct H as [H1 H2]; simpl in *; rewrite andb_true_iff in H2;
          destruct H2 as [H2 H3]; rewrite andb_true_iff in *;
            destruct H3 as [H3 _ ];
              unfold isLengthOf in H3; 
                apply beq_nat_true in H3; rewrite H3;
                  split; [ assumption | idtac ];
                    rewrite forallb_forall in *; intros l0 H.
                      apply in_app_or in H; destruct H. 
                      apply In_rev in H; apply H2; assumption.
                      simpl in H; destruct H; 
                            [ rewrite H; apply lengthIsLengthOf | elim H].
Qed.

Lemma tailOfMatrixIsMatrix:
  forall (A:Type) (m:list(list A)),
  isMatrix m = true -> isMatrix(tail m) = true.
Proof.
  intros A m H; destruct m.
    auto.
    unfold isMatrix in H. rewrite andb_true_iff in H. destruct  H. simpl in H.
      destruct(length l).
        discriminate H.
        apply (sameLengthsSuccIsMatrix m n); assumption.
Qed.

Lemma isMatrixApp: 
  forall (A:Type) (m1 m2 : list(list A)),
    isMatrix(m1 ++ m2) = true -> isMatrix m1 = true /\ isMatrix m2 = true.
Proof.
  intros A m1 m2 H;split;
    destruct m1 ; destruct m2; simpl in * ; auto.
  rewrite <- (app_nil_end m1) in H; assumption.
  rewrite andb_true_iff in H; destruct H as [H H1]; rewrite H ; simpl;
    rewrite forallbApp in H1; rewrite andb_true_iff in H1; destruct  H1; 
      assumption.
  rewrite andb_true_iff in H; destruct H as [H H1];
    rewrite forallbApp in H1; rewrite andb_true_iff in H1; 
      destruct  H1 as [H1 H2] ; simpl in H2;
        rewrite andb_true_iff in H2; destruct H2 as [H2 H3].
  assert(length l0 = length l) as Hlx0.
      apply beq_nat_true; unfold isLengthOf in H2; apply H2; apply in_or_app; intuition.
      rewrite Hlx0. 
    rewrite H. simpl ; assumption. 
Qed.

(** * Type of a matrix *)

Definition matrix (A:Type) := {m:list (list A) | isMatrix m = true}.

(** ** The width and length of a matrix *)
Program Definition 
  matrixSize (A:Type) (m:matrix A) :=
  match m with 
    | nil => (0,0)
    | h::t => (List.length m, List.length h)
  end.

Lemma matrixSizeIsProofIrrelevant:
  forall (A:Type)(m1 m2 :matrix A),
  proj1_sig m1 = proj1_sig m2 ->
  matrixSize m1 = matrixSize m2.
Proof.
  intros A m1 m2 H.
  destruct m1 as [m1 Hm1].   
  destruct m2 as [m2 Hm2].
  simpl in H.
  unfold matrixSize ; subst ; reflexivity. 
Qed.

Program Definition strongHead (A:Type)(l:{l:list A | l<>nil}) : 
  {a:A | forall d, hd d l = a}.
Proof.
  intros A l; destruct l as [l Hl];
    destruct l; 
      [ contradict Hl ; reflexivity | 
        exists a; reflexivity ].
Defined.

Program Definition head (A:Type)(l:list A)(H:l<>nil): A :=
  proj1_sig(strongHead (exist _ l H)).

Program Lemma matrixSizeNotZero: 
  forall(A:Type)(m:matrix A)(H:proj1_sig m <> nil),
    matrixSize m = (List.length m, List.length (head H)).
Proof.
  intros A m H;
    unfold matrixSize; 
      destruct m as [m Hm];
        destruct m.
          simpl in *; contradict H; reflexivity.
          assert(length (head H) = length l) by
            (unfold head, strongHead ; simpl; reflexivity);
            simpl in *; subst ; reflexivity.
Qed.

Program Lemma inMatrixLength: 
  forall (A:Type) (m:matrix A),
  forall l, In l m -> List.length l = snd(matrixSize m).
Proof.
  intros A m l Hl.
  destruct m as [m Hm].
  destruct m.
    elim Hl.
    simpl in * ; destruct Hl as [Hl | Hl].
      subst ; auto.
      rewrite andb_true_iff in Hm ; destruct Hm as [Hm1 Hm];
        rewrite forallb_forall in Hm ; unfold isLengthOf in Hm. 
      rewrite (beq_nat_true _ _ (Hm l Hl)) ; reflexivity.
Qed.

(** * Square matrices *)  

Program Definition isSquareMatrix (A:Type) (m:matrix A) := 
  let size := matrixSize m in 
    beq_nat (fst size)(snd size).

Definition squareMatrix (A:Type) := 
  {m:matrix A | isSquareMatrix m = true}.

Program Lemma inSquareMatrix (A:Type) (m:squareMatrix A): 
    forall l, In l m -> List.length l = fst(matrixSize m).
Proof.
  intros A m l H.
  destruct m as [m Hm1]; rewrite (inMatrixLength m _ H);
    unfold isSquareMatrix in * ; simpl in * ;
      apply beq_nat_true in Hm1; 
        symmetry ; assumption.
Qed.

(** * In the type [strongMatrix] the proof that all columns have the
   same length is attached to each column. *)

Definition strongMatrix (A:Type)(n:nat) := list {l:list A | length l = n }.

Program Definition 
  strongMatrixOfMatrix (A:Type)(m:matrix A) : 
  strongMatrix A (snd(matrixSize m)) := 
  inSig  _ m (inMatrixLength m).  

Lemma strongMatrixOfMatrixIsProofIrrelevant:
  forall (A:Type)(m1 m2:matrix A),
  proj1_sig  m1 = proj1_sig  m2 ->
  listProj (strongMatrixOfMatrix m1) = listProj (strongMatrixOfMatrix m2).
Proof.
  intros A m1 m2 H; 
    unfold strongMatrixOfMatrix;
      rewrite inSigProj; rewrite inSigProj;
        assumption.
Qed.

(** * Transposition of a matrix *)

Program Definition 
  getRow (A:Type)(n:nat)(m:strongMatrix A n)(i:boundedNat n) : 
  { l : list A | length l = length m } := 
  List.map (fun l =>boundedStrongNth l i) m.
Next Obligation.
  induction m ; auto ; simpl in *.
  rewrite IHm ; reflexivity.
Defined.

Lemma getRowIsProofIrrelevant:
  forall
  (A:Type)(n:nat)(m1 m2:strongMatrix A n)(i1 i2:boundedNat n),
  listProj  m1 = listProj  m2 ->
  proj1_sig i1 = proj1_sig i2 ->
  proj1_sig (getRow m1 i1) = 
  proj1_sig (getRow m2 i2).
Proof.
  intros A n m1 m2 i1 i2 H H0.
  simpl in *.
  generalize dependent m2 ;
  induction m1 ; intros m2 H ;destruct m2 ; auto ; try(discriminate H).
  simpl in *.
  inversion H.
  rewrite boundedStrongNthIsProofIrrelevant with (l2:=s)(pos2:=i2).
  rewrite IHm1 with (m2:=m2). reflexivity.
  assumption.
  reflexivity.
  assumption.
  assumption.
Qed.

Program Lemma getRowLength : 
  forall (A:Type)(n:nat)(m:strongMatrix A n)(i:boundedNat n),
    List.length(getRow m i) = List.length m.
Proof.
  intros A n m i; destruct(getRow m i); auto.
Qed.

Hint Rewrite getRowLength: length.
  
Program Definition transpose (A:Type) (m:matrix A): matrix A :=
  let n := snd(matrixSize m) in
    let f := fun (i:boundedNat n) => 
      getRow(strongMatrixOfMatrix m) i in
    List.map f (strongSeq 0 n).
Next Obligation.
  destruct m as [m Hm] ; destruct m. 
    auto.
    destruct l.
      auto.
      simpl in *. rewrite forallb_forall. intros e He.
      rewrite map_length. 
      rewrite in_map_iff in He. destruct He.
      destruct H. rewrite <- H. simpl.
      unfold isLengthOf. simpl. 
      rewrite map_length.
      symmetry.
      apply beq_nat_refl.
Defined.
    

Lemma transposeIsProofIrrelevant:
  forall (A:Type)(m1 m2: matrix A),
    proj1_sig m1 = proj1_sig m2 ->
    proj1_sig (transpose m1) = proj1_sig (transpose m2).
Proof.
  intros A m1 m2 H;
    assert(matrixSize m1 = matrixSize m2) as HH.
      apply (matrixSizeIsProofIrrelevant m1 m2 H).
    unfold transpose ; simpl in *; apply mapProj1SigExt.
    intros x1 x2 H0;
      apply mapProj1SigExt.
        intros x0 x3 H1;
          destruct x1, x2;
            apply boundedStrongNthIsProofIrrelevant.
              rewrite HH ; reflexivity.
              assumption.
              assumption.
        apply strongMatrixOfMatrixIsProofIrrelevant; assumption.
      rewrite inSigProj ; rewrite inSigProj ; rewrite HH ; reflexivity.
Qed.

Program Lemma transposeNil : 
  forall (A:Type)(m:matrix A),
    proj1_sig (transpose m) = nil -> proj1_sig m = nil.
Proof.
  intros A m H;
    destruct m as [m Hm];
      destruct m;
        [ unfold transpose ; reflexivity | idtac].
  unfold transpose in *; simpl in *.
  rewrite <-lengthNilEquiv in H.
  autorewrite with length in H.
  apply andb_true_iff in Hm; destruct Hm as [Hm1 Hm2]; contradict Hm1;
    rewrite H; simpl; auto.
Qed.

Definition swap (A B: Type) (x:A*B) : B*A := (snd x, fst x).

Lemma transposeMatrixSize: 
  forall  (A:Type) (mat:matrix A),
  matrixSize(transpose mat) = swap(matrixSize mat).
Proof.
  intros A mat;
  destruct mat as [mat Hmat]; destruct mat.
    unfold swap, matrixSize; reflexivity.
    set(mat':=(exist (fun m : list (list A) => isMatrix m = true)(l :: mat) Hmat));
      assert(proj1_sig mat'<>nil) as Hmat' by (intro H; discriminate);
        assert(proj1_sig(transpose mat')<>nil) as Htmat' by
          (intro H ; apply Hmat' ; apply transposeNil ; assumption).
    case_eq (transpose mat');intros tm Htm Heq; destruct tm.
      rewrite Heq in Htmat'; contradict Htmat'; reflexivity.
      unfold matrixSize, swap; simpl.
      assert(proj1_sig (transpose mat') = l0::tm) by (rewrite Heq ; reflexivity).
      assert(length(proj1_sig(transpose mat')) = length l) by 
        (unfold transpose, strongSeq; autorewrite with length;
          rewrite inSigLength with 
            (P:=(fun n : nat => n < 0 + snd (matrixSize mat')));
            autorewrite with length; unfold matrixSize, mat'; reflexivity).
      rewrite H in H0 ; simpl in H0; rewrite H0;
        unfold transpose, strongSeq in H; simpl in H.
      assert(In l0 (l0::tm)) as Hinl0 by intuition.
      rewrite <- H in Hinl0;
        apply in_map_iff in Hinl0;
          destruct Hinl0 as [pos Hpos]; destruct Hpos.
      assert(length l0 = S(length mat)) by 
        (rewrite <- H1; simpl; autorewrite with length;
          rewrite inSigLength with (P:=(fun l1 : list A => length l1 = length l));
            reflexivity).
      rewrite H3; reflexivity.
Qed.

Lemma transposedMatrixHeight:
  forall(A:Type) (m:matrix A),
  fst(matrixSize m)  = snd (matrixSize (transpose m)).
Proof.
  intros A m.
  rewrite transposeMatrixSize.
  destruct(matrixSize(transpose m)) ; simpl; reflexivity.
Qed.
  

Lemma transposedMatrixWidth:
  forall (A:Type) (m:matrix A),
  snd(matrixSize m)  = fst (matrixSize (transpose m)).
Proof.
  intros A m.
  rewrite transposeMatrixSize.
  destruct(matrixSize(transpose m)) ; simpl; reflexivity.
Qed.


Lemma squareMatrixLengthTranspose:
  forall A  (m:matrix A), 
    isSquareMatrix m = true ->
    List.length (proj1_sig m) = 
    List.length (proj1_sig (transpose m)).
Proof.
  assert(forall A (m:matrix A), List.length (proj1_sig m) = fst(matrixSize m)) as Hl by
    (intros A' m'; unfold matrixSize; destruct (proj1_sig m'); auto).
  unfold isSquareMatrix; intros A m H; symmetry in H; apply beq_nat_eq in H;
    repeat rewrite Hl; rewrite H; apply transposedMatrixWidth.
Qed.

(*
Program Lemma inMatrix :
  forall (A:Type)(m:matrix A)(l:list A),
  In l m ->
  exists n, l = getRow (strongMatrixOfMatrix(transpose m)) n.
Proof.
  intros A m l H.
  destruct m as [m Hm]. simpl in H. 
  induction m.
    contradict H.
    destruct H as [H | H].
    (* Case 0 *)
    set(P:=fun n=> n <
      snd
      (matrixSize
        (transpose
          (exist (fun m : list (list A) => isMatrix m = true) (a :: m) Hm)))).
    assert(P 0) as P0 
      by (unfold P; rewrite <- transposedMatrixHeight;
        unfold matrixSize; simpl; apply lt_0_Sn).
    exists (exist P 0 P0).
    unfold transpose.  simpl.
    simpl.
    assert(isMatrix m = true) as Hm' 
      by (eapply sameLengthsNotZeroIsMatrix; apply Hm).
    destruct (IHm Hm') as [n Hn].
    simpl.
*)          
    

Program Definition get (A:Type)(m:matrix A) 
  (column:boundedNat(fst(matrixSize m))) 
  (row:boundedNat(snd(matrixSize m))) : A := 
  strongNth (strongNth m column) row.
Next Obligation.
  destruct m as [m Hm].
  destruct m ; simpl in *.
    destruct column; absurd(x<0);[ omega | assumption ].
    destruct column; assumption.
Qed.
Next Obligation.
  destruct m as [m Hm].
  destruct m.
    destruct column; absurd(x<0);[ omega | assumption ].
    destruct row; destruct column; simpl in *;
      destruct x0 ; [ assumption | idtac ].
    apply lt_S_n in l1;
      apply nth_In with (d:=l) in l1;
        rewrite andb_true_iff in Hm ; destruct Hm as [Hm1 Hm] ;
          rewrite forallb_forall in Hm.
    assert(length (nth x0 m l) = length l) as H.
      apply beq_nat_true; apply Hm; assumption.
    rewrite H ; assumption.
Qed.

(*
Program Lemma transposeIsCorrect 
  (A:Type)(m:matrix A):
  forall 
    (i:boundedNat (fst(matrixSize m))) 
    (j:boundedNat (snd(matrixSize m))),
    let i' := boundedNatConversion (fst(matrixSize m)) (snd(matrixSize(transpose m))) i (transposedMatrixHeight m) in 
    let j' := boundedNatConversion (snd(matrixSize m)) (fst(matrixSize(transpose m))) j (transposedMatrixWidth m) in 
    get m i j = get (transpose m) j' i'.
Proof.
Qed.
*)